/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package vcl

import (
	"regexp"
	"text/template"
)

const selfShardTmplSrc = `
import std;
import directors;
import blob;
import blobdigest;
import taskvar;

probe vk8s_probe_varnish {
	.request = "HEAD /vk8s_cluster_health HTTP/1.1"
	           "Host: vk8s_cluster"
	           "Connection: close";
{{- if .Probe.Timeout}}
	.timeout = {{.Probe.Timeout}};
{{- end}}
{{- if .Probe.Interval}}
	.interval = {{.Probe.Interval}};
{{- end}}
{{- if .Probe.Initial}}
	.initial = {{.Probe.Initial}};
{{- end}}
{{- if .Probe.Window}}
	.window = {{.Probe.Window}};
{{- end}}
{{- if .Probe.Threshold}}
	.threshold = {{.Probe.Threshold}};
{{- end}}
}

{{range $node := .Nodes -}}
backend {{nodeName $node}} {
	.host = "{{(index $node.Addresses 0).IP}}";
	.port = "{{(index $node.Addresses 0).Port}}";
	.probe = vk8s_probe_varnish;
}

{{end -}}

acl vk8s_cluster_acl {
{{- range $node := .Nodes}}
	"{{(index $node.Addresses 0).IP}}";
{{- end}}
}

sub vcl_init {
	new vk8s_cluster_param = directors.shard_param();
	new vk8s_cluster = directors.shard();
	vk8s_cluster.associate(vk8s_cluster_param.use());
	{{range $node := .Nodes -}}
	vk8s_cluster.add_backend({{nodeName $node}});
	{{end -}}
	vk8s_cluster.reconfigure();
	{{- digest_init . }}
	{{ if .PrimaryOnly -}}
	new vk8s_cluster_primary = taskvar.backend();
	{{- end }}
}

sub vcl_recv {
	{{ if .PrimaryOnly -}}
	{{- digest_update 'c' . }}
	vk8s_cluster_primary.set(vk8s_cluster.backend(resolve=NOW
		{{- key 'c' .}}));
	if (remote.ip !~ vk8s_cluster_acl
	    && "" + vk8s_cluster_primary.get() != server.identity) {
		set req.backend_hint = vk8s_cluster_primary.get();
		return (pipe);
	}
	else {{ end }}if (remote.ip ~ vk8s_cluster_acl) {
		if (req.http.Host == "vk8s_cluster") {
			if (req.url == "/vk8s_cluster_health") {
				return (synth(200));
			}
			return (synth(404));
		}

		# prevent deadlock for accidental cyclic requests
		set req.hash_ignore_busy = true;

		# if we're async, don't deliver stale
		if (req.http.VK8S-Is-Bgfetch == "true") {
			set req.grace = 0s;
		}

		return (hash);
	}
}

sub vcl_backend_fetch {
	{{- digest_update 'b' . }}
	vk8s_cluster_param.set({{ key 'b' .}});
	if (bereq.retries == 0 
	    && !bereq.uncacheable
	    && remote.ip !~ vk8s_cluster_acl
	    && "" + vk8s_cluster.backend(resolve=NOW) != server.identity) {
		set bereq.backend = vk8s_cluster.backend(resolve=LAZY);
		set bereq.http.VK8S-Is-Bgfetch = bereq.is_bgfetch;
		return (fetch);
	}
}

sub vcl_backend_response {
	if (bereq.backend == vk8s_cluster.backend(resolve=LAZY)) {
	        if (beresp.http.VK8S-Cluster-TTL) {
		        set beresp.ttl = std.duration(
                	    beresp.http.VK8S-Cluster-TTL + "s", 1s);
		        if (beresp.ttl > {{.MaxSecondaryTTL}}) {
			        set beresp.ttl = {{.MaxSecondaryTTL}};
		        }
		        unset beresp.http.VK8S-Cluster-TTL;
	        }
                else {
			set beresp.uncacheable = true;
		}
		return (deliver);
	}
}

sub vcl_backend_error {
	if (bereq.backend == vk8s_cluster.backend(resolve=LAZY)) {
		return (deliver);
	}
}

sub vcl_deliver {
	unset resp.http.VK8S-Cluster-TTL;
	if (remote.ip ~ vk8s_cluster_acl
		{{- if .PrimaryOnly }} && ! vk8s_cluster_primary.defined()
		{{- end }}) {
		if (! obj.uncacheable) {
			set resp.http.VK8S-Cluster-TTL = obj.ttl;
		}
		return (deliver);
	}
}
`

var reqMatch = regexp.MustCompile("^req")

func context(ctx rune, key string) string {
	if ctx == 'c' {
		return key
	}
	return reqMatch.ReplaceAllLiteralString(key, "bereq")
}

func keyParams(ctx rune, shard ShardCluster) string {
	pfx := ""
	if ctx == 'c' {
		pfx = ", "
	}
	switch shard.By {
	case ByHash:
		return ""
	case URL:
		return pfx + "by=URL"
	case Key:
		return pfx + "by=KEY, key=vk8s_cluster.key(" +
			context(ctx, shard.Key) + ")"
	}
	return pfx + "by=BLOB, key_blob=vk8s_shard_digest.final()"
}

func digestInit(shard ShardCluster) string {
	if shard.By != Blob {
		return ""
	}
	return "\n\tnew vk8s_shard_digest = blobdigest.digest(" +
		shard.Algo.String() + ");"
}

func digestUpdate(ctx rune, shard ShardCluster) string {
	if shard.By != Blob {
		return ""
	}
	return "\n\tvk8s_shard_digest.update(blob.decode(encoded=" +
		context(ctx, shard.Key) + "));"
}

const selfShardName = "self-sharding"

var shardFuncMap = template.FuncMap{
	"key":           keyParams,
	"digest_init":   digestInit,
	"digest_update": digestUpdate,
	"nodeName": func(svc Service) string {
		// MUST match -i setting in the varnishd invocaton
		// (server.identity)
		addr := svc.Addresses[0]
		return "vk8s_" + addr.PodNamespace + "_" + addr.PodName
	},
}

var shardTmpl = template.Must(template.New(selfShardName).Funcs(shardFuncMap).
	Parse(selfShardTmplSrc))

// ResetShardTmpl sets the VCL template for the VarnishConfig
// shard feature to its current "official" value. Invoked on
// TemplateConfig deletion, only needed when devmode is activated for
// the controller.
func ResetShardTmpl() {
	shardTmpl = template.Must(template.New(selfShardName).
		Funcs(shardFuncMap).Parse(selfShardTmplSrc))
}

// SetShardTmpl parses src as a text/template, using the FuncMap
// defined for the shard template, which is used to generate VCL for
// the VarnishConfig shard feature. On success, the shard template
// is replaced. If the parse fails, then the error is returned and the
// shard template is unchanged.
//
// Only used when devmode is activated for the controller.
func SetShardTmpl(src string) error {
	newTmpl, err := template.New(selfShardName).Funcs(shardFuncMap).
		Parse(src)
	if err != nil {
		return err
	}
	shardTmpl = newTmpl
	return nil
}
